import { isEqual } from "lodash";
import { logger } from "../../logger";

import {
    IWorker,
    IWorkerCompletedGroupTask,
    IWorkerDatabase,
} from "../../interfaces/IWorker";
import type { SetOptional } from "type-fest";
import { FingerprintGenerator } from "fingerprint-generator";
import { IFaucetMail } from "../../interfaces/IFaucetMail";
import { JSONDatabase, SQLDatabase } from "../ADatabase";
import { JSONFaucetMails } from "../JSONDatabases/FaucetMail";


export const db = new (class SQLWorkerDatabase extends SQLDatabase {
    constructor() {
        super("db.sqlite");
        this.rawDB.schema.hasTable("workers").then((result) => {
            if (result != true) {
                this.rawDB.schema
                    .createTable("workers", (table) => {
                        table
                            .string("phoneNumber", 12)
                            .notNullable()
                            .primary()
                            .unique(),
                            table.integer("telegramID").notNullable().unique(),
                            table.bigInteger("apiID").notNullable(),
                            table.string("apiHash").notNullable(),
                            table.string("password").notNullable(),
                            table.jsonb("faucetMail").notNullable().unique(),
                            table
                                .string("stringSession")
                                .notNullable()
                                .unique(),
                            table.float("balance").notNullable().defaultTo(0.0),
                            table.boolean("canBeRefferal").notNullable(),
                            table
                                .jsonb("browserFingerprint")
                                .notNullable()
                                .unique()
                                .comment(
                                    "I save it in json data, so as not to create another table"
                                ),
                            table.integer("refferal").defaultTo(null),
                            table
                                .boolean("troubled")
                                .notNullable()
                                .comment("if worker has maximum groups");
                    })
                    .then(() => {
                        logger.debug(
                            "Created `workers` table into `db.sqlite`"
                        );
                    });
            }
        });
        this.rawDB.schema.hasTable("completedGroupsTasks").then((result) => {
            if (result != true) {
                this.rawDB.schema
                    .createTable("completedGroupsTasks", (table) => {
                        table.increments("id").primary().unique(),
                            table.string("worker", 12).notNullable(),
                            table.string("groupID").notNullable(),
                            table.timestamp("timeToLeave").notNullable();
                    })
                    .then(() => {
                        logger.debug(
                            "Created `completedGroupsTasks` table into `db.sqlite`"
                        );
                    });
            }
        });
    }
    convertWorker(worker: IWorker): IWorkerDatabase {
        let convertedWorker: IWorkerDatabase = structuredClone(
            worker
        ) as unknown as IWorkerDatabase;
        Object.keys(worker).forEach(function (key, index) {
            if (worker[key] === Object(worker[key])) {
                convertedWorker[key] = JSON.stringify(worker[key]);
            } else {
                convertedWorker[key] = worker[key];
            }
        });
        delete convertedWorker["completedGroupsTasks"];
        delete convertedWorker["withdraws"];
        return convertedWorker;
    }
    async getCompletedGroups(
        worker: Pick<IWorker, "phoneNumber">
    ): Promise<IWorkerCompletedGroupTask[]> {
        return await this.rawDB("completedGroupsTasks")
            .where({ worker: worker.phoneNumber })
            .select("timeToLeave", "groupID");
    }
    async getWorkers(): Promise<IWorker[]> {
        try {
            return Promise.all(
                (
                    await this.rawDB.select().from<IWorkerDatabase>("workers")
                ).map(async (worker) => {
                    Object.keys(worker).forEach(function (key, index) {
                        try {
                            worker[key] = JSON.parse(worker[key]);
                        } catch (e) {
                            // normal prop
                        }
                    });
                    worker["withdraws"] = [];
                    const completedGroupsTasks = await this.getCompletedGroups(
                        worker
                    );
                    worker["completedGroupsTasks"] = completedGroupsTasks;

                    return worker as unknown as IWorker;
                })
            );
        } catch {
            return [];
        }
    }
    async findRefferal(maxRefferals: number): Promise<number | null> {
        let refferal: number | null = null;
        const workers = await this.getWorkers();
        if (workers.length > 0) {
            for (const mainWorker of workers) {
                let counter = 0;
                if (mainWorker.canBeRefferal == true) {
                    const _workers = workers.filter((worker, workerIndex) => {
                        return workerIndex !== workers.indexOf(mainWorker);
                    });
                    for (const _worker of _workers) {
                        if (_worker.refferal == mainWorker.telegramID) {
                            counter += 1;
                        }
                    }

                    if (counter < maxRefferals) {
                        refferal = Number(mainWorker.telegramID);
                    }
                }
            }
        }
        return refferal;
    }

    async addCompletedGroups(
        worker: Pick<IWorker, "phoneNumber">,
        group: IWorkerCompletedGroupTask
    ) {
        await this.rawDB("completedGroupsTasks").insert({
            worker: worker.phoneNumber,
            groupID: group.groupID,
            timeToLeave: group.timeToLeave,
        });
    }

    async addUser(
        worker: Omit<
            IWorker,
            | "balance"
            | "withdraws"
            | "completedGroupsTasks"
            | "canBeRefferal"
            | "browserFingerprint"
            | "faucetMail"
            | "troubled"
        >,
        wallets_json: string
    ): Promise<void> {
        const fingerprintGenerator = new FingerprintGenerator();
        const fingerprint = fingerprintGenerator.getFingerprint({
            devices: ["desktop"],
            browsers: ["chrome"],
        }).fingerprint;

        const faucet: IFaucetMail | undefined = await this.findWallet(
            wallets_json
        );
        if (faucet === undefined) {
            logger.error("Add new faucet accounts");
            throw new Error("Add new faucet accounts");
        } else {
            const canBeRefferal: boolean = Math.random() < 0.5;

            const _worker: IWorker = {
                phoneNumber: worker.phoneNumber,
                telegramID: worker.telegramID,
                apiID: worker.apiID,
                apiHash: worker.apiHash,
                password: worker.password,
                faucetMail: faucet,
                stringSession: worker.stringSession,
                balance: 0.0,
                withdraws: [],
                completedGroupsTasks: [],
                canBeRefferal: canBeRefferal,
                refferal: worker.refferal,
                browserFingerprint: fingerprint,
                troubled: false,
            };
            await this.rawDB("workers").insert(this.convertWorker(_worker));
        }
    }
    async findWallet(walletsJson: string): Promise<IFaucetMail | undefined> {
        const walletsJsonDB = new JSONFaucetMails(walletsJson);
        await walletsJsonDB.load();
        const wallets = await walletsJsonDB.getWallets();
        if (Object.keys(wallets).length == 0) {
            logger.error("File which you choose doesnt have anything");
            throw new Error("File which you choose doesnt have anything");
        }
        const workers = await this.getWorkers();
        let faucetWallet: IFaucetMail | undefined = undefined;
        for (const wallet in wallets) {
            let flag = false;
            if (workers.length > 0) {
                for (const worker of workers) {
                    if (worker.faucetMail.mail.address == wallet) {
                        flag = true;
                        break;
                    }
                }
            }
            if (flag == false) {
                faucetWallet = {
                    faucet: {
                        username: wallets[wallet].faucet.username,
                        password: wallets[wallet].faucet.password,
                    },
                    mail: {
                        address: wallet,
                        password: wallets[wallet].mail.password,
                    },
                };
                break;
            }
        }
        return faucetWallet;
    }
})();